/* Use Gnulib. */
# include <config.h>

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <stdbool.h>

#include <easy-machine/easy-machine-generate-code.h>

#include <easy-machine/easy-machine.h>
#include <easy-machine/easy-machine-utility.h>
#include <easy-machine/easy-machine-meta-instructions.h>


/* Code generation: files.
 * ************************************************************************** */

/* It would be annoying to pass a lot of identical FILE * arguments to many
   functions; I can factor them all into a struct. */
struct easy_machine_files
{
  /* Each field is a descriptor, open for writing, for a file named like
     the field. */
  FILE *machine_h;
  FILE *instructions_h;
  FILE *instruction_names_c;
  FILE *interpreter_c;
  FILE *assembler_c;
  FILE *text;
};

/* A conveniece macro for fprintf'ing to a file, assumuing that the struct
   above is passed via pointer thru a formal parameter named "files". */
/* Perform a fprintf call exiting fatally in case of error. */
#define EMIT(...)                                               \
  do                                                            \
    {                                                           \
      if (fprintf (__VA_ARGS__) < 0)                            \
        easy_machine_fatal ("could not write to output file");  \
    }                                                           \
  while (false)




/* Code generation: generate instructions from meta-instructions.
 * ************************************************************************** */

static void
easy_machine_generate_variant (struct easy_machine_files *files,
                               const struct easy_machine_instruction_descriptor *mi,
                               enum easy_machine_instruction_argument_kind *kinds,
                               easy_machine_instruction_opcode opcode)
{
  char argument_pattern [EASY_MACHINE_MAX_ARGS + 10];
  int argument_pattern_next_index = 0;
  EMIT (files->instructions_h, "#define EASY_MACHINE_INSTRUCTION_OPCODE_%s", mi->mnemonic);
  printf ("opcode %i: %s ", (int) opcode, mi->mnemonic);
  int i;
  for (i = 0; i < mi->arg_no; i ++)
    switch (kinds [i])
      {
      case easy_machine_instruction_argument_kind_immediate:
        argument_pattern [argument_pattern_next_index ++] = 'i';
        printf ("i ");
        break;
      case easy_machine_instruction_argument_kind_register:
        argument_pattern [argument_pattern_next_index ++] = 'r';
        printf ("r ");
        break;
      case easy_machine_instruction_argument_kind_terminator:
        easy_machine_fatal ("this should never happen: terminator\n");
      default:
        easy_machine_fatal ("this should never happen: %i\n", (int) kinds [i]);
      };
  argument_pattern [argument_pattern_next_index] = '\0';
  EMIT (files->instructions_h, "_%s  ", argument_pattern);
  EMIT (files->instructions_h, "%i\n", (int) opcode);

  EMIT (files->instruction_names_c, "      \"%s_%s\",\n", mi->mnemonic,
        argument_pattern);

  /* Handle the interpreter. */
  EMIT (files->interpreter_c, "#define INSTRUCTION_NAME \"%s_%s\"\n",
        mi->mnemonic, argument_pattern);
  for (i = 0; i < mi->arg_no; i ++)
    if (kinds [i] == easy_machine_instruction_argument_kind_immediate)
      EMIT (files->interpreter_c, "#define ARG%i ((easy_machine_word)(ip->arg%i))\n", i, i);
    else // register kind
      EMIT (files->interpreter_c, "#define ARG%i (s->registers [ip->arg%i])\n", i, i);
  EMIT (files->interpreter_c, "    case EASY_MACHINE_INSTRUCTION_OPCODE_%s_%s:\n", mi->mnemonic, argument_pattern);
  EMIT (files->interpreter_c, "      {\n");
  EMIT (files->interpreter_c, "        %s\n", mi->source);
  EMIT (files->interpreter_c, "        break;\n");
  EMIT (files->interpreter_c, "      }\n");
  for (i = 0; i < mi->arg_no; i ++)
    EMIT (files->interpreter_c, "#undef ARG%i\n", i);
  EMIT (files->interpreter_c, "#undef INSTRUCTION_NAME\n");
  EMIT (files->interpreter_c, "\n");

  printf ("\n");
  printf ("  (%s)", mi->synopsis);
  printf ("\n");
  printf ("  %s", mi->description);
  printf ("\n\n");

  /* Handle the assembler part. */
  EMIT (files->assembler_c, "  if (! strcmp (i->mnemonic, \"%s\"))\n", mi->mnemonic);
  EMIT (files->assembler_c, "    mnemonic_recognized = true;\n");
  EMIT (files->assembler_c, "  if (! strcmp (i->mnemonic, \"%s\")\n", mi->mnemonic);
  EMIT (files->assembler_c, "      && i->arg_no != %i)\n", (int) mi->arg_no);
  EMIT (files->assembler_c, "    easy_machine_asm_fatal (\"invalid operand number for %s: given %%i, required %i\", (int) i->arg_no);\n", mi->mnemonic, (int) mi->arg_no);
  EMIT (files->assembler_c, "  if (! strcmp (i->mnemonic, \"%s\")", mi->mnemonic);
  for (i = 0; i < mi->arg_no; i ++)
    {
      char *kind_name;
      switch (kinds [i])
        {
        case easy_machine_instruction_argument_kind_immediate:
          kind_name = "easy_machine_asm_instruction_argument_kind_numeric";
          break;
        case easy_machine_instruction_argument_kind_register:
          kind_name = "easy_machine_asm_instruction_argument_kind_register";
          break;
        case easy_machine_instruction_argument_kind_terminator:
          easy_machine_fatal ("B this should never happen: terminator\n");
        default:
          easy_machine_fatal ("B this should never happen: %i\n", (int) kinds [i]);
        };
      EMIT (files->assembler_c, "\n      && i->args [%i].kind == %s", (int) i, kind_name);
    }
  EMIT (files->assembler_c, ")\n");
  EMIT (files->assembler_c, "    {\n");
  EMIT (files->assembler_c, "      i->opcode = EASY_MACHINE_INSTRUCTION_OPCODE_%s_%s;\n", mi->mnemonic, argument_pattern);
  EMIT (files->assembler_c, "      return;\n");
  EMIT (files->assembler_c, "    }\n");

  /* Handle the documentation. */
  EMIT (files->text, "%s_%s (opcode %i decimal, 0x%x hexadecimal)\n", mi->mnemonic, argument_pattern, (int) opcode, (unsigned) opcode);
  EMIT (files->text, "Synopsis: %s\n", mi->synopsis);
  EMIT (files->text, "%s\n", mi->description);
  EMIT (files->text, "----------------------------------------------------------\n");
  EMIT (files->text, "\n");
}

static void
easy_machine_generate_variants (struct easy_machine_files *files,
                                const struct easy_machine_instruction_descriptor *mi,
                                int next_arg_index,
                                enum easy_machine_instruction_argument_kind *kinds,
                                easy_machine_instruction_opcode *next_opcode)
{
  if (next_arg_index == mi->arg_no)
    {
      /* We have completed a new variant.  Generate code for it. */
      easy_machine_generate_variant (files, mi, kinds, * next_opcode);

      /* The next variatn we generate will have a new opcode. */
      (* next_opcode) ++;
      return;
    }

  enum easy_machine_instruction_argument_kind k;
  for (k = 1; k < easy_machine_instruction_argument_kind_terminator; k *= 2)
    if (mi->args [next_arg_index] & k)
      {
        kinds [next_arg_index] = k;
        easy_machine_generate_variants (files, mi, next_arg_index + 1, kinds, next_opcode);
      }
}

static void
easy_machine_generate_interpreter_opening (struct easy_machine_files *files)
{
  EMIT (files->instructions_h, "/* This file is machine-generated. */\n");
  EMIT (files->interpreter_c, "/* Run the instruction at the current instruction pointer.\n");
  EMIT (files->interpreter_c, "   Return the address of the next instruction to execute. */\n");
  EMIT (files->interpreter_c, "easy_machine_word\n");
  EMIT (files->interpreter_c, "easy_machine_interpret_instruction (struct easy_machine_thread_state *s, struct easy_machine_assmbler_command_line *cl)\n");
  EMIT (files->interpreter_c, "{\n");
  EMIT (files->interpreter_c, "  easy_machine_word vip = s->vpc;\n");
  EMIT (files->interpreter_c, "  if (cl->debug) printf (\"0x%%lx: executing \", (unsigned long) vip);\n");
  EMIT (files->interpreter_c, "  struct easy_machine_instruction *ip\n");
  EMIT (files->interpreter_c, "    = ((struct easy_machine_instruction *)\n");
  EMIT (files->interpreter_c, "       easy_machine_virtual_address_to_pointer (vip));\n");
  EMIT (files->interpreter_c, "  easy_machine_instruction_opcode opcode = ip->opcode;\n");
  EMIT (files->interpreter_c, "  if (opcode < 0 || opcode >= EASY_MACHINE_INSTRUCTION_NO)\n");
  EMIT (files->interpreter_c, "    {\n");
  EMIT (files->interpreter_c, "      if (cl->debug) printf (\"<invalid>\\n\");\n");
  EMIT (files->interpreter_c, "      easy_machine_fatal (\"invalid opcode exception: opcode %%li\", (long) opcode);\n");
  EMIT (files->interpreter_c, "    }\n");
  EMIT (files->interpreter_c, "  if (cl->debug) printf (\"%%s\\n\", easy_machine_instruction_names [opcode]);\n");
  EMIT (files->interpreter_c, "  switch (opcode)\n");
  EMIT (files->interpreter_c, "    {\n");
}

static void
easy_machine_generate_assembler_opening (struct easy_machine_files *files)
{
  EMIT (files->assembler_c, "/* This file is machine-generated. */\n");
  EMIT (files->assembler_c, "void\n");
  EMIT (files->assembler_c, "easy_machine_asm_instruction_set_opcode (struct easy_machine_routine *r, struct easy_machine_asm_instruction *i)\n");
  EMIT (files->assembler_c, "{\n");
  EMIT (files->assembler_c, "  bool mnemonic_recognized = false;\n");
}
static void
easy_machine_generate_assembler_closing (struct easy_machine_files *files)
{
  EMIT (files->assembler_c, "  if (! mnemonic_recognized)\n");
  EMIT (files->assembler_c, "    easy_machine_asm_fatal (\"invalid instruction mnemonic %%s\", i->mnemonic);\n");
  EMIT (files->assembler_c, "  easy_machine_asm_fatal (\"invalid parameter kinds for %%s\", i->mnemonic);\n");
  EMIT (files->assembler_c, "}\n");
}

static void
easy_machine_generate_interpreter_closing (struct easy_machine_files *files)
{
  EMIT (files->interpreter_c, "    default:\n");
  EMIT (files->interpreter_c, "      easy_machine_fatal (\"invalid opcode %%lu\", (unsigned long) opcode);\n");
  EMIT (files->interpreter_c, "    }\n");
  EMIT (files->interpreter_c, "  /* Fall thru the next instruction if we didn't return already. */\n");
  EMIT (files->interpreter_c, "  return vip + sizeof (struct easy_machine_instruction);\n");
  EMIT (files->interpreter_c, "}\n");
}

static void
easy_machine_generate_openings (struct easy_machine_files *files)
{
  EMIT (files->instructions_h, "/* This file is machine-generated. */\n");

  EMIT (files->instruction_names_c, "/* This file is machine-generated. */\n");
  EMIT (files->instruction_names_c, "const char\n");
  EMIT (files->instruction_names_c, "*easy_machine_instruction_names []\n");
  EMIT (files->instruction_names_c, "  = {\n");

  easy_machine_generate_interpreter_opening (files);
  easy_machine_generate_assembler_opening (files);

  EMIT (files->text, "# This is a quick specification of the instrutions.\n");
}

static void
easy_machine_generate_closings (struct easy_machine_files *files,
                                easy_machine_instruction_opcode next_opcode)
{
  easy_machine_generate_assembler_closing (files);
  easy_machine_generate_interpreter_closing (files);

  EMIT (files->instructions_h, "#define EASY_MACHINE_INSTRUCTION_NO  %i\n\n",
        (int) next_opcode);
  EMIT (files->instructions_h, "#ifndef EASY_MACHINE_INSTRUCTIONS_H_\n");
  EMIT (files->instructions_h, "#endif // #ifndef EASY_MACHINE_INSTRUCTIONS_H_\n");

  EMIT (files->instruction_names_c, "    };\n");
}

static void
easy_machine_generate_instructions (struct easy_machine_files *files)
{
  /* Generate the opening, fixed part of files. */
  easy_machine_generate_openings (files);

  /* Keep an array of argument kinds, which will be set in every valid way for
     every meta-instruction, this way generating every possible instruction
     variant.  Keep a counter tracking which opcode will be the next. */
  easy_machine_instruction_opcode next_opcode = 0;
  enum easy_machine_instruction_argument_kind kinds [EASY_MACHINE_MAX_ARGS];
  int i;
  /* For every meta-instruction... */
  for (i = 0; i < easy_machine_meta_instruction_descriptor_no; i ++)
    {
      const struct easy_machine_instruction_descriptor *mi
        = easy_machine_meta_instruction_descriptors + i;
      if (mi->arg_no > EASY_MACHINE_MAX_ARGS)
        easy_machine_fatal ("this should never happen: too many arguments for %s\n",
                            mi->mnemonic);
      /* Generate every variant.  This will write to kinds multiple times,
         always reusing the same space. */
      easy_machine_generate_variants (files, mi, 0, kinds, & next_opcode);
    }

  /* Generate the trailing, end part of files. */
  easy_machine_generate_closings (files, next_opcode);
}




/* Code generation: entry point.
 * ************************************************************************** */

void
easy_machine_generate_code_do_work (const struct command_line *command_line)
{
  /* Open every file. */
  struct easy_machine_files files;
#define EASY_MACHINE_OPEN(field, pathname)                               \
  do                                                                     \
    {                                                                    \
      files.field = fopen (pathname, "w");                               \
      if (files.field == NULL)                                           \
        easy_machine_fatal ("could not open %s for writing", pathname);  \
    }                                                                    \
  while (false)
  EASY_MACHINE_OPEN (machine_h, "/tmp/machine.h");
  EASY_MACHINE_OPEN (instructions_h, "easy-machine/easy-machine-instructions-machine-generated.h");
  EASY_MACHINE_OPEN (instruction_names_c, "easy-machine/easy-machine-machine-instruction-names-machine-generated.c");
  EASY_MACHINE_OPEN (interpreter_c, "easy-machine/easy-machine-machine-interpreter-machine-generated.c");
  EASY_MACHINE_OPEN (assembler_c, "easy-machine/easy-machine-machine-assembler-machine-generated.c");
  EASY_MACHINE_OPEN (text, "Documentation");
#undef EASY_MACHINE_OPEN

  /* Generate code starting from the meta-instructions. */
  easy_machine_generate_instructions (& files);

  /* Close files. */
  fclose (files.machine_h);
  fclose (files.instructions_h);
  fclose (files.instruction_names_c);
  fclose (files.interpreter_c);
  fclose (files.assembler_c);
  fclose (files.text);
}
