/* Easy Machine.

   Instruction definitions. */

/* Use Gnulib. */
#include <config.h>

#include <stdarg.h>

#include <easy-machine/easy-machine-meta-instructions.h>
#include <easy-machine/easy-machine-utility.h>


/* Hand-written meta-instruction specifications.
 * ************************************************************************** */

/* This is the hand-written specification for every meta-instruction.  The code
   generator handles the mapping from this to actual instructions, including
   opcode numbers. */
const struct easy_machine_instruction_descriptor
easy_machine_meta_instruction_descriptors []
= {
   /* Special instructions. */
   {.mnemonic = "trap",
    .synopsis = "Trap",
    .description = "Trap, killing the process",
    .arg_no = 0,
    .source = "easy_machine_exception (\"trap on user request\");"},
   {.mnemonic = "syscall",
    .synopsis = "Invoke a kernel operation",
    .description = "Call the kernel procedure specified in the first argument,"
    " passing the other arguments as the kernel call arguments.  Some syscalls"
    " modify registers to return result from the kernel to the user process.",
    .arg_no = 3,
    .args = {easy_machine_instruction_argument_kind_immediate,
             (easy_machine_instruction_argument_kind_immediate
              | easy_machine_instruction_argument_kind_register),
             (easy_machine_instruction_argument_kind_immediate
              | easy_machine_instruction_argument_kind_register)},
    .source = "easy_machine_syscall (s, ARG0, ARG1, ARG2, cl);"},
   /* Do-nothing instructions. */
   {.mnemonic = "nop",
    .synopsis = "Do nothing",
    .description = "Do absolutely nothing.",
    .arg_no = 0,
    .source = ""},
   {.mnemonic = "nopi",
    .description = "Do nothing, but take an immediate argument.  This is useful"
    " for debugging, so that when disassembling the human user can find"
    " specific program points.",
    .arg_no = 1,
    .args = {easy_machine_instruction_argument_kind_immediate},
    .source = ""},
   /* Instructions for loading constants. */
   {.mnemonic = "mov",
    .synopsis = "Set a register to a constant value or to the content of another register",
    .description = "Set the register given in the first argument to the"
    " constant or the value of the register given in the second argument.",
    .arg_no = 2,
    .args = {easy_machine_instruction_argument_kind_register,
             (easy_machine_instruction_argument_kind_register
              | easy_machine_instruction_argument_kind_immediate)},
    .source = "ARG0 = ARG1;"},
   /* Arithmetic instructions. */
   {.mnemonic = "add",
    .synopsis = "Sum two integer numbers",
    .description = "Sum the second and third argument, writing the result into"
    " the first argument.",
    .arg_no = 3,
    .args = {easy_machine_instruction_argument_kind_register,
             (easy_machine_instruction_argument_kind_immediate
              | easy_machine_instruction_argument_kind_register),
             (easy_machine_instruction_argument_kind_immediate
              | easy_machine_instruction_argument_kind_register)},
    .source = "ARG0 = ARG1 + ARG2;"},
   {.mnemonic = "sub",
    .synopsis = "Subtract two integer numbers",
    .description = "Subtract the third argument from the second, writing the result"
    " into the first argument.",
    .arg_no = 3,
    .args = {easy_machine_instruction_argument_kind_register,
             (easy_machine_instruction_argument_kind_immediate
              | easy_machine_instruction_argument_kind_register),
             (easy_machine_instruction_argument_kind_immediate
              | easy_machine_instruction_argument_kind_register)},
    .source = "ARG0 = ARG1 - ARG2;"},
   {.mnemonic = "mul",
    .synopsis = "Multiply two integer numbers",
    .description = "Multiply the second and third arguments, writing the result"
    " into the first argument.",
    .arg_no = 3,
    .args = {easy_machine_instruction_argument_kind_register,
             (easy_machine_instruction_argument_kind_immediate
              | easy_machine_instruction_argument_kind_register),
             (easy_machine_instruction_argument_kind_immediate
              | easy_machine_instruction_argument_kind_register)},
    .source = "ARG0 = ARG1 * ARG2;"},
   {.mnemonic = "div",
    .synopsis = "Integer division",
    .description = "Divide the second argument by the third, writing the"
    " quotient into the first argument.  Do not fail if even if the second"
    " argument is zero: in that case, just set the result register to an"
    " unspecified value.",
    .arg_no = 3,
    .args = {easy_machine_instruction_argument_kind_register,
             (easy_machine_instruction_argument_kind_immediate
              | easy_machine_instruction_argument_kind_register),
             (easy_machine_instruction_argument_kind_immediate
              | easy_machine_instruction_argument_kind_register)},
    .source = "if (ARG2 == 0) easy_machine_exception (\"division by zero\"); ARG0 = ARG1 / ARG2;"},
   {.mnemonic = "mod",
    .synopsis = "Compute the remainder of the division between two integer"
    " numbers",
    .description = "Compute the remainder of the division of the second argument"
    " by the third, writing the result into the first argument.  Do not fail if"
    " even if the second argument is zero: in that case, just set the result"
    " register to an unspecified value.  This only computes the remainder, not"
    " the quotient.",
    .arg_no = 3,
    .args = {easy_machine_instruction_argument_kind_register,
             (easy_machine_instruction_argument_kind_immediate
              | easy_machine_instruction_argument_kind_register),
             (easy_machine_instruction_argument_kind_immediate
              | easy_machine_instruction_argument_kind_register)},
    .source = "if (ARG2 == 0) easy_machine_exception (\"remainder of division by zero\"); ARG0 = ARG1 % ARG2;"},
   /* Memory access instructions. */
   {.mnemonic = "lw",
    .synopsis = "Load a 32-bit word from memory",
    .description = "Load a 32-bit word from the memory address obtained by the"
    " sum of the second and third argument; write the result into the first"
    " argument.  There are no alignment requirements.",
    .arg_no = 3,
    .args = {easy_machine_instruction_argument_kind_register,
             (easy_machine_instruction_argument_kind_immediate
              | easy_machine_instruction_argument_kind_register),
             (easy_machine_instruction_argument_kind_immediate
              | easy_machine_instruction_argument_kind_register)},
    .source = "ARG0 = easy_machine_load_word (ARG1 + ARG2);"},
   {.mnemonic = "sw",
    .synopsis = "Store a 32-bit word into memory",
    .description = "Store a 32-bit word from the first argument into the memory"
    " address obtained by the sum of the second and third arguments."
    "  There are no alignment requirements.",
    .arg_no = 3,
    .args = {(easy_machine_instruction_argument_kind_immediate
              | easy_machine_instruction_argument_kind_register),
             (easy_machine_instruction_argument_kind_immediate
              | easy_machine_instruction_argument_kind_register),
             (easy_machine_instruction_argument_kind_immediate
              | easy_machine_instruction_argument_kind_register)},
    .source = "easy_machine_store_word (ARG1 + ARG2, ARG0);"},
   {.mnemonic = "lbz",
    .synopsis = "Load an 8-bit byte from memory, zero-extending",
    .description = "Load a zero-extended byte from the memory address obtained"
    " by the sum of the second and third argument; write the result into the"
    " first argument.  There are no alignment restrictions.",
    .arg_no = 3,
    .args = {easy_machine_instruction_argument_kind_register,
             (easy_machine_instruction_argument_kind_immediate
              | easy_machine_instruction_argument_kind_register),
             (easy_machine_instruction_argument_kind_immediate
              | easy_machine_instruction_argument_kind_register)},
    .source = "ARG0 = easy_machine_load_byte (ARG1 + ARG2);"},
   /*
   {.mnemonic = "lbs",
    .synopsis = "Load an 8-bit byte from memory, sign-extending",
    .description = "Load a sign-extended byte from the memory address obtained"
    " by the sum of the second and third argument; write the result into the"
    " first argument.  There are no alignment restrictions.",
    .arg_no = 3,
    .args = {easy_machine_instruction_argument_kind_register,
             (easy_machine_instruction_argument_kind_immediate
              | easy_machine_instruction_argument_kind_register),
             (easy_machine_instruction_argument_kind_immediate
              | easy_machine_instruction_argument_kind_register)},
    .source = "easy_machine_fatal (\"unimplemented\" INSTRUCTION_NAME);"},
   */
   {.mnemonic = "sb",
    .synopsis = "Store an 8-bit byte into memory",
    .description = "Store the lowest byte of the first argument into the memory"
    " address obtained by the sum of the second and third arguments."
    "  There are no alignment requirements.",
    .arg_no = 3,
    .args = {(easy_machine_instruction_argument_kind_immediate
              | easy_machine_instruction_argument_kind_register),
             (easy_machine_instruction_argument_kind_immediate
              | easy_machine_instruction_argument_kind_register),
             (easy_machine_instruction_argument_kind_immediate
              | easy_machine_instruction_argument_kind_register)},
    .source = "easy_machine_store_byte (ARG1 + ARG2, ARG0);"},
   /* Unconditional branching instructions. */
   {.mnemonic = "ja",
    .synopsis = "Jump, absolute",
    .description = "Jump to the instruction at the absolute address given"
    " in the argument.",
    .arg_no = 1,
    .args = {(easy_machine_instruction_argument_kind_immediate
              | easy_machine_instruction_argument_kind_register)},
    .source = "return ARG0;"},
   {.mnemonic = "jr",
    .synopsis = "Jump, PC-relative",
    .description = "Jump to the instruction at the PC-relative offset given"
    " in the argument.  The offset is computed considering the beginning of"
    " the jumping instruction as the base.",
    .arg_no = 1,
    .args = {(easy_machine_instruction_argument_kind_immediate
              | easy_machine_instruction_argument_kind_register)},
    .source = "return vip + ARG0;"},
   {.mnemonic = "jrl",
    .synopsis = "Jump and link, PC-relative",
    .description = "Jump to the instruction at the PC-relative offset given"
    " in the second argument.  The offset is computed considering the beginning"
    " of the jumping instruction as the base.  Load the address immediately"
    " following the jumping instruction to the register specified in the first"
    " argument.",
    .arg_no = 2,
    .args = {easy_machine_instruction_argument_kind_register,
             (easy_machine_instruction_argument_kind_immediate
              | easy_machine_instruction_argument_kind_register)},
    .source = "ARG0 = vip + sizeof (struct easy_machine_instruction); return vip + ARG1;"},
   {.mnemonic = "jal",
    .synopsis = "Jump and link, absolute",
    .description = "Jump to the instruction at the absolute address given"
    " in the second argument.  Load the address immediately following the"
    " jumping instruction to the register specified in the first"
    " argument.",
    .arg_no = 2,
    .args = {easy_machine_instruction_argument_kind_register,
             (easy_machine_instruction_argument_kind_immediate
              | easy_machine_instruction_argument_kind_register)},
    .source = "ARG0 = vip + sizeof (struct easy_machine_instruction); return ARG1;"},
   /* Conditional branching instructions, PC-relative. */
   {.mnemonic = "jeqr",
    .synopsis = "Jump on equal, PC-relative",
    .description = "If the first argument is equal to the second argument then"
    " jump to the instruction at the PC-relative address (from the beginning of"
    " this instruction) given in the third argument.  Do nothing otherwise.",
    .arg_no = 3,
    .args = {(easy_machine_instruction_argument_kind_immediate
              | easy_machine_instruction_argument_kind_register),
             (easy_machine_instruction_argument_kind_immediate
              | easy_machine_instruction_argument_kind_register),
             easy_machine_instruction_argument_kind_immediate},
    .source = "if (ARG0 == ARG1) return vip + ARG2;"},
   {.mnemonic = "jner",
    .synopsis = "Jump on non-equal, PC-relative",
    .description = "If the first argument is different from the second argument"
    " then jump to the instruction at the PC-relative address (from the"
    " beginning of this instruction) given in the third argument.  Do nothing"
    " otherwise.",
    .arg_no = 3,
    .args = {(easy_machine_instruction_argument_kind_immediate
              | easy_machine_instruction_argument_kind_register),
             (easy_machine_instruction_argument_kind_immediate
              | easy_machine_instruction_argument_kind_register),
             easy_machine_instruction_argument_kind_immediate},
    .source = "if (ARG0 != ARG1) return vip + ARG2;"},
   {.mnemonic = "jltsr",
    .synopsis = "Jump on less, signed, PC-relative",
    .description = "If the first argument is less then the second argument"
    " (in a signed comparison) then jump to the instruction at the PC-relative"
    " address (from the beginning of this instruction) given in the third"
    " argument.  Do nothing otherwise.",
    .arg_no = 3,
    .args = {(easy_machine_instruction_argument_kind_immediate
              | easy_machine_instruction_argument_kind_register),
             (easy_machine_instruction_argument_kind_immediate
              | easy_machine_instruction_argument_kind_register),
             easy_machine_instruction_argument_kind_immediate},
    .source = "if ((easy_machine_word)ARG0 < (easy_machine_word)ARG1) return vip + ARG2;"},
   {.mnemonic = "jgtsr",
    .synopsis = "Jump on greater, signed, PC-relative",
    .description = "If the first argument is greater then the second argument"
    " (in a signed comparison) then jump to the instruction at the PC-relative"
    " address (from the beginning of this instruction) given in the third"
    " argument.  Do nothing otherwise.",
    .arg_no = 3,
    .args = {(easy_machine_instruction_argument_kind_immediate
              | easy_machine_instruction_argument_kind_register),
             (easy_machine_instruction_argument_kind_immediate
              | easy_machine_instruction_argument_kind_register),
             easy_machine_instruction_argument_kind_immediate},
    .source = "if ((easy_machine_word)ARG0 > (easy_machine_word)ARG1) return vip + ARG2;"},
   {.mnemonic = "jltur",
    .synopsis = "Jump on less, unsigned, PC-relative",
    .description = "If the first argument is less then the second argument"
    " (in a unsigned comparison) then jump to the instruction at the PC-relative"
    " address (from the beginning of this instruction) given in the third"
    " argument.  Do nothing otherwise.",
    .arg_no = 3,
    .args = {(easy_machine_instruction_argument_kind_immediate
              | easy_machine_instruction_argument_kind_register),
             (easy_machine_instruction_argument_kind_immediate
              | easy_machine_instruction_argument_kind_register),
             easy_machine_instruction_argument_kind_immediate},
    .source = "if ((easy_machine_uword)ARG0 < (easy_machine_uword)ARG1) return vip + ARG2;"},
   {.mnemonic = "jgtur",
    .synopsis = "Jump on greater, unsigned, PC-relative",
    .description = "If the first argument is greater then the second argument"
    " (in a unsigned comparison) then jump to the instruction at the PC-relative"
    " address (from the beginning of this instruction) given in the third"
    " argument.  Do nothing otherwise.",
    .arg_no = 3,
    .args = {(easy_machine_instruction_argument_kind_immediate
              | easy_machine_instruction_argument_kind_register),
             (easy_machine_instruction_argument_kind_immediate
              | easy_machine_instruction_argument_kind_register),
             easy_machine_instruction_argument_kind_immediate},
    .source = "if ((easy_machine_uword)ARG0 > (easy_machine_uword)ARG1) return vip + ARG2;"},
   {.mnemonic = "jlesr",
    .synopsis = "Jump on less or equal, signed, PC-relative",
    .description = "If the first argument is less then the second argument"
    " (in a signed comparison) then jump to the instruction at the PC-relative"
    " address (from the beginning of this instruction) given in the third"
    " argument.  Do nothing otherwise.",
    .arg_no = 3,
    .args = {(easy_machine_instruction_argument_kind_immediate
              | easy_machine_instruction_argument_kind_register),
             (easy_machine_instruction_argument_kind_immediate
              | easy_machine_instruction_argument_kind_register),
             easy_machine_instruction_argument_kind_immediate},
    .source = "if ((easy_machine_word)ARG0 <= (easy_machine_word)ARG1) return vip + ARG2;"},
   {.mnemonic = "jgesr",
    .synopsis = "Jump on greater or equal, signed, PC-relative",
    .description = "If the first argument is greater then the second argument"
    " (in a signed comparison) then jump to the instruction at the PC-relative"
    " address (from the beginning of this instruction) given in the third"
    " argument.  Do nothing otherwise.",
    .arg_no = 3,
    .args = {(easy_machine_instruction_argument_kind_immediate
              | easy_machine_instruction_argument_kind_register),
             (easy_machine_instruction_argument_kind_immediate
              | easy_machine_instruction_argument_kind_register),
             easy_machine_instruction_argument_kind_immediate},
    .source = "if ((easy_machine_word)ARG0 >= (easy_machine_word)ARG1) return vip + ARG2;"},
   {.mnemonic = "jleur",
    .synopsis = "Jump on less or equal, unsigned, PC-relative",
    .description = "If the first argument is less then the second argument"
    " (in a unsigned comparison) then jump to the instruction at the PC-relative"
    " address (from the beginning of this instruction) given in the third"
    " argument.  Do nothing otherwise.",
    .arg_no = 3,
    .args = {(easy_machine_instruction_argument_kind_immediate
              | easy_machine_instruction_argument_kind_register),
             (easy_machine_instruction_argument_kind_immediate
              | easy_machine_instruction_argument_kind_register),
             easy_machine_instruction_argument_kind_immediate},
    .source = "if ((easy_machine_uword)ARG0 <= (easy_machine_uword)ARG1) return vip + ARG2;"},
   {.mnemonic = "jgeur",
    .synopsis = "Jump on greater or equal, unsigned, PC-relative",
    .description = "If the first argument is greater then the second argument"
    " (in a unsigned comparison) then jump to the instruction at the PC-relative"
    " address (from the beginning of this instruction) given in the third"
    " argument.  Do nothing otherwise.",
    .arg_no = 3,
    .args = {(easy_machine_instruction_argument_kind_immediate
              | easy_machine_instruction_argument_kind_register),
             (easy_machine_instruction_argument_kind_immediate
              | easy_machine_instruction_argument_kind_register),
             easy_machine_instruction_argument_kind_immediate},
    .source = "if ((easy_machine_uword)ARG0 >= (easy_machine_uword)ARG1) return vip + ARG2;"},
   /* Absolute branches. */
   {.mnemonic = "jeqa",
    .synopsis = "Jump on equal, absolute",
    .description = "If the first argument is equal to the second argument then"
    " jump to the instruction at the absolute address given in the third"
    " argument.  Do nothing otherwise.",
    .arg_no = 3,
    .args = {(easy_machine_instruction_argument_kind_immediate
              | easy_machine_instruction_argument_kind_register),
             (easy_machine_instruction_argument_kind_immediate
              | easy_machine_instruction_argument_kind_register),
             easy_machine_instruction_argument_kind_immediate},
    .source = "if (ARG0 == ARG1) return ARG2;"},
   {.mnemonic = "jnea",
    .synopsis = "Jump on non-equal, absolute",
    .description = "If the first argument is different from the second argument"
    " then jump to the instruction at the absolute address given in the third"
    " argument.  Do nothing otherwise.",
    .arg_no = 3,
    .args = {(easy_machine_instruction_argument_kind_immediate
              | easy_machine_instruction_argument_kind_register),
             (easy_machine_instruction_argument_kind_immediate
              | easy_machine_instruction_argument_kind_register),
             easy_machine_instruction_argument_kind_immediate},
    .source = "if (ARG0 != ARG1) return ARG2;"},
   {.mnemonic = "jltsa",
    .synopsis = "Jump on less, signed, absolute",
    .description = "If the first argument is less then the second argument"
    " (in a signed comparison) then jump to the instruction at the absolute"
    " addressgiven in the third argument.  Do nothing otherwise.",
    .arg_no = 3,
    .args = {(easy_machine_instruction_argument_kind_immediate
              | easy_machine_instruction_argument_kind_register),
             (easy_machine_instruction_argument_kind_immediate
              | easy_machine_instruction_argument_kind_register),
             easy_machine_instruction_argument_kind_immediate},
    .source = "if ((easy_machine_word)ARG0 < (easy_machine_word)ARG1) return ARG2;"},
   {.mnemonic = "jgtsa",
    .synopsis = "Jump on greater, signed, absolute",
    .description = "If the first argument is greater then the second argument"
    " (in a signed comparison) then jump to the instruction at the absolute"
    " address given in the third argument.  Do nothing otherwise.",
    .arg_no = 3,
    .args = {(easy_machine_instruction_argument_kind_immediate
              | easy_machine_instruction_argument_kind_register),
             (easy_machine_instruction_argument_kind_immediate
              | easy_machine_instruction_argument_kind_register),
             easy_machine_instruction_argument_kind_immediate},
    .source = "if ((easy_machine_word)ARG0 > (easy_machine_word)ARG1) return ARG2;"},
   {.mnemonic = "jltua",
    .synopsis = "Jump on less, unsigned, absolute",
    .description = "If the first argument is less then the second argument"
    " (in a unsigned comparison) then jump to the instruction at the absolute"
    " address given in the third argument.  Do nothing otherwise.",
    .arg_no = 3,
    .args = {(easy_machine_instruction_argument_kind_immediate
              | easy_machine_instruction_argument_kind_register),
             (easy_machine_instruction_argument_kind_immediate
              | easy_machine_instruction_argument_kind_register),
             easy_machine_instruction_argument_kind_immediate},
    .source = "if ((easy_machine_uword)ARG0 < (easy_machine_uword)ARG1) return ARG2;"},
   {.mnemonic = "jgtua",
    .synopsis = "Jump on greater, unsigned, absolute",
    .description = "If the first argument is greater then the second argument"
    " (in a unsigned comparison) then jump to the instruction at the absolute"
    " address given in the third argument.  Do nothing otherwise.",
    .arg_no = 3,
    .args = {(easy_machine_instruction_argument_kind_immediate
              | easy_machine_instruction_argument_kind_register),
             (easy_machine_instruction_argument_kind_immediate
              | easy_machine_instruction_argument_kind_register),
             easy_machine_instruction_argument_kind_immediate},
    .source = "if ((easy_machine_uword)ARG0 > (easy_machine_uword)ARG1) return ARG2;"},
   {.mnemonic = "jlesa",
    .synopsis = "Jump on less or equal, signed, absolute",
    .description = "If the first argument is less then the second argument"
    " (in a signed comparison) then jump to the instruction at the absolute"
    " address given in the third argument.  Do nothing otherwise.",
    .arg_no = 3,
    .args = {(easy_machine_instruction_argument_kind_immediate
              | easy_machine_instruction_argument_kind_register),
             (easy_machine_instruction_argument_kind_immediate
              | easy_machine_instruction_argument_kind_register),
             easy_machine_instruction_argument_kind_immediate},
    .source = "if ((easy_machine_word)ARG0 <= (easy_machine_word)ARG1) return ARG2;"},
   {.mnemonic = "jgesa",
    .synopsis = "Jump on greater or equal, signed, absolute",
    .description = "If the first argument is greater then the second argument"
    " (in a signed comparison) then jump to the instruction at the absolute"
    " address given in the third argument.  Do nothing otherwise.",
    .arg_no = 3,
    .args = {(easy_machine_instruction_argument_kind_immediate
              | easy_machine_instruction_argument_kind_register),
             (easy_machine_instruction_argument_kind_immediate
              | easy_machine_instruction_argument_kind_register),
             easy_machine_instruction_argument_kind_immediate},
    .source = "if ((easy_machine_word)ARG0 >= (easy_machine_word)ARG1) return ARG2;"},
   {.mnemonic = "jleua",
    .synopsis = "Jump on less or equal, unsigned, absolute",
    .description = "If the first argument is less then the second argument"
    " (in a unsigned comparison) then jump to the instruction at the absolute"
    " address given in the third argument.  Do nothing otherwise.",
    .arg_no = 3,
    .args = {(easy_machine_instruction_argument_kind_immediate
              | easy_machine_instruction_argument_kind_register),
             (easy_machine_instruction_argument_kind_immediate
              | easy_machine_instruction_argument_kind_register),
             easy_machine_instruction_argument_kind_immediate},
    .source = "if ((easy_machine_uword)ARG0 <= (easy_machine_uword)ARG1) return ARG2;"},
   {.mnemonic = "jgeua",
    .synopsis = "Jump on greater or equal, unsigned, absolute",
    .description = "If the first argument is greater then the second argument"
    " (in a unsigned comparison) then jump to the instruction at the absolute"
    " address given in the third argument.  Do nothing otherwise.",
    .arg_no = 3,
    .args = {(easy_machine_instruction_argument_kind_immediate
              | easy_machine_instruction_argument_kind_register),
             (easy_machine_instruction_argument_kind_immediate
              | easy_machine_instruction_argument_kind_register),
             easy_machine_instruction_argument_kind_immediate},
    .source = "if ((easy_machine_uword)ARG0 >= (easy_machine_uword)ARG1) return ARG2;"},
   /* Stack instructions. */
   {.mnemonic = "push",
    .synopsis = "Push a word on the stack",
    .description = "Store the argument to the memory location pointed by %r255, then"
                   " increment %r255 by four to make the stack pointer refer a fresh"
                   " location.",
    .arg_no = 1,
    .args = {(easy_machine_instruction_argument_kind_immediate
              | easy_machine_instruction_argument_kind_register)},
    .source = "easy_machine_store_word (s->registers [255], ARG0); s->registers [255] += 4;"},
   {.mnemonic = "pop",
    .synopsis = "Pop a word from the stack",
    .description = "Decrement %r255 by four, to make the stack pointer refer to the"
                   " location which was last pushed to; then copy the stack top to the"
                   " register given in the argument.",
    .arg_no = 1,
    .args = {easy_machine_instruction_argument_kind_register},
    .source = "s->registers [255] -= 4; ARG0 = easy_machine_load_word (s->registers [255]);"},
   {.mnemonic = "updatestack",
    .synopsis = "Reserve or release a given number of bytes on the stack",
    .description = "Set the register in the first argument to the address of the current stack"
    " top (which should be currently unused); increment the stack pointer in %r255 by the"
    " quantity in bytes given in the second argument.  If the second argument is positive this"
    " instruction reserves space on the stack; if it is negative, it releases it.",
    .arg_no = 2,
    .args = {easy_machine_instruction_argument_kind_register,
             (easy_machine_instruction_argument_kind_immediate
              | easy_machine_instruction_argument_kind_register)},
    .source = "ARG0 = s->registers [255]; s->registers [255] += ARG1;"}
   /* End of instruction specifications. */
  };

const size_t
easy_machine_meta_instruction_descriptor_no
  = (sizeof (easy_machine_meta_instruction_descriptors)
     / sizeof (struct easy_machine_instruction_descriptor));
